#ifndef Matrix_h
#define Matrix_h

#include <stddef.h>
#include <initializer_list>
#include <type_traits>
#include <cmath>

extern void * ___enabler ;

/**
 * M×N行列クラス。
 *
 * @param MSize         行の数。
 * @param NSize         列の数。
 * @param T             要素の型。
 */
template <size_t MSize, size_t NSize, class T = float>
class Matrix
{
public :
  static const size_t M = MSize ;
  static const size_t N = NSize ;

private :
  /**
   * 要素。
   */
  T m_elm[M * N] ;

public :
  /**
   * m_elm[] を初期化しないコンストラクタ。
   */
  Matrix() = default ;

  /**
   * コピーコンストラクタ。
   *
   * @param U       コピー元行列の要素型。
   * @param another コピー元行列
   */
  template <class U>
  Matrix(const Matrix<M, N, U>& another)
  {
    T *p = m_elm ;
    const U *q = (const U*)another ;
    for (T *pEnd = p + M*N ; p < pEnd ; ) *p++ = (T)*q++ ;
  }

  /**
   * 配列で値を指定して初期化。
   */
  Matrix(const T (&mat)[M][N])
  {
    for (size_t m = 0 ; m < M ; ++m) {
      for (size_t n = 0 ; n < N ; ++n) {
        m_elm[m + M * n] = mat[m][n] ;
      }
    }
  }

  /**
   * 初期化リストを用いた行列の初期化。
   *
   * @param value M*Nコの引数列
   */
  template <class ... U, typename std::enable_if<sizeof...(U)==M*N>::type *& = ___enabler>
  explicit Matrix(const U& ... value)
  {
    auto it = std::initializer_list<T>({(T)value ...}).begin() ;

    for (size_t m = 0 ; m < M ; ++m) {
      for (size_t n = 0 ; n < N ; ++n) {
        m_elm[m + M * n] = *it++ ;
      }
    }
  }

  /**
   * 転置行列を求める。
   *
   * @return 転置行列。
   */
  Matrix<N, M, T> transpose() const
  {
    Matrix<N, M, T> t ;
    T *p = (T*)t ;

    for (size_t m = 0 ; m < M ; ++m) {
      for (size_t n = 0 ; n < N ; ++n) {
        *p++ = m_elm[m + M * n] ;
      }
    }

    return t ;
  }

  /**
   * 要素の取得／設定。
   *
   * @param  m 行
   * @param  n 列
   * @return   m行n列の要素の値
   */
  T& at(size_t m, size_t n) { return m_elm[m + M * n] ; }
  const T& at(size_t m, size_t n) const { return m_elm[m + M * n] ; }

  /**
   * 要素配列へのポインタを取得。
   *
   * @return 要素(0,0)へのポインタ。
   */
  explicit operator const T *() const { return &m_elm[0]; }
  explicit operator T *() { return &m_elm[0]; }
  
  /**
   * ゼロ行列。
   */
  static const Matrix O ;
  

  // 以下、正方行列専用
public :
  /**
   * 行列k*Iを初期化。
   *
   * @param k
   */
  Matrix(const T& k) ;
} ;


template <size_t MSize, size_t NSize, class T>
const Matrix<MSize, NSize, T> Matrix<MSize, NSize, T>::O = {{0}} ;

template <size_t MSize, size_t NSize, class T>
Matrix<MSize, NSize, T>::Matrix(const T& t)
{
  static_assert(M==N, "Matrix::Matrix(const T&) for non-square matrix.") ;

  T *p = (T*)*this ;
  T *pEnd = p + M*M ;
  for (int i = 0 ; p < pEnd ; ++i, ++p) *p = ((i % (M+1)) == 0 ? t : (T)0) ;
}

// 4x4行列の場合、Matrix4<型名>と書けるようにする
template <class T = float>
using Matrix4 = Matrix<4, 4, T> ;

// 行ベクトル／列ベクトルの場合、ColVector<次元数, 型名>, Vector<次元数, 型名>と書けるようにする
template <size_t NSize, class T = float>
using ColVector = Matrix<1, NSize, T> ;
template <size_t MSize, class T = float>
using Vector = Matrix<MSize, 1, T> ;

// さらに、4x4のfloat行列の場合、4x4のdouble配列の場合の別名を作る
typedef Matrix4<float> Matrix4f ;
typedef Matrix4<double> Matrix4d ;

#endif
