#include "Context.h"
#include "TextureObject.h"

std::vector<gl::Context*> gl::Context::m_contexts ;

gl::Context::Context()
{
  m_contexts.push_back(this) ;
}

gl::Context::~Context()
{
  auto it = std::find(m_contexts.begin(), m_contexts.end(), this) ;
  if (it != m_contexts.end()) {
    if (it != m_contexts.end() - 1) {
      // 一番最後の要素を抜けた場所に持ってくる
      *it = m_contexts.back() ;
    }
    // リサイズ
    m_contexts.pop_back() ;
  }

  // 不要とマークされたGLオブジェクトを全て削除する
  for (GLObject *obj : m_erased_objects) {
    obj->suppress_gl_delete() ;
    delete obj ;
  }
  // 既存のテクスチャオブジェクトを全て削除する
  for (auto pair : m_texture_map) {
    pair.second->suppress_gl_delete() ;
    delete pair.second ;
  }
}

void gl::Context::awake()
{
  // コンテキストが有効になった際、不要とマークされたGLオブジェクトを全て削除する
  if (!m_erased_objects.empty()) {
    for (GLObject *obj : m_erased_objects) {
      delete obj ;
    }
    m_erased_objects.clear() ;
  }
}


/**
 * テクスチャをbindするためのオブジェクトを取得。
 */
gl::TextureRenderer gl::Context::bind(gl::Texture *tex)
{
  glEnable(static_cast<GLenum>(tex->getTarget())) ;

  TextureObject *obj = findObject(tex) ;
  if (!obj) obj = createObject(tex) ;

  obj->update() ;

  return TextureRenderer(obj) ;
}

/**
 * テクスチャから、コンテキストごとのテクスチャオブジェクトを取得。
 */
gl::TextureObject *gl::Context::findObject(gl::Texture *tex)
{
  auto it = m_texture_map.find(tex) ;
  if (it == m_texture_map.end()) return nullptr ;

  return it->second ;
}

/**
 * テクスチャから、コンテキストごとのテクスチャオブジェクトを作成。
 */
gl::TextureObject *gl::Context::createObject(gl::Texture *tex)
{
  TextureObject *obj = new TextureObject(tex) ;
  obj->update() ;
  
  m_texture_map.insert(std::make_pair(tex, obj)) ;

  return obj ;
}

/**
 * テクスチャから、コンテキストごとのテクスチャオブジェクトを削除。
 */
void gl::Context::eraseObject(gl::Texture *tex)
{
  auto it = m_texture_map.find(tex) ;
  if (it == m_texture_map.end()) return ;

  // 削除
  m_texture_map.erase(it) ;

  // GLオブジェクトの削除予約
  m_erased_objects.push_back(it->second) ;
}

