#ifndef gl__Matrix_h
#define gl__Matrix_h

#include "../noncopyable.h"
#include "../glall.h"
#include "../Matrix.h"
#include <stdexcept>

namespace gl
{
  template <class T>
  struct MatrixFunctions {} ;

  template <>
  struct MatrixFunctions<GLfloat>
  {
    inline static void set(const ::Matrix<4, 4, GLfloat> &mat) { glLoadMatrixf((const GLfloat*)mat) ; }
    inline static void get(GLenum pname, ::Matrix<4, 4, GLfloat> &mat) { glGetFloatv(pname, (GLfloat*)mat) ; }
    inline static void mult(const ::Matrix<4, 4, GLfloat> &mat) { glMultMatrixf((const GLfloat*)mat) ; }
  } ;
  
  template <>
  struct MatrixFunctions<GLdouble>
  {
    inline static void set(const ::Matrix<4, 4, GLdouble> &mat) { glLoadMatrixd((const GLdouble*)mat) ; }
    inline static void get(GLenum pname, ::Matrix<4, 4, GLdouble> &mat) { glGetDoublev(pname, (GLdouble*)mat) ; }
    inline static void mult(const ::Matrix<4, 4, GLdouble> &mat) { glMultMatrixd((const GLdouble*)mat) ; }
  } ;


  /**
   * 一時的に行列スタックを変更するためのクラス。
   *
   * @param SetterConst glMatrixMode()に指定する定数。
   * @param GetterConst glGet*()に指定する定数。
   */
  template <GLenum SetterConst, GLenum GetterConst>
  class Matrix : noncopyable<Matrix<SetterConst, GetterConst>>
  {
    // gl::Contextクラスでのみインスタンス化可能。
    friend class Context ;

  private :
    /**
     * 現在値をスタックに保存。
     */
    Matrix()
    {
      // matrix modeを変更
      GLint mode ;
      glGetIntegerv(GL_MATRIX_MODE, &mode) ;
      if (mode != SetterConst) glMatrixMode(SetterConst) ;

      // 現在値を退避
      glPushMatrix() ;
      if (GL_STACK_OVERFLOW == glGetError()) {
        throw std::overflow_error("matrix stack overflow.") ;
      }

      // matri modeを元に戻す
      if (mode != SetterConst) glMatrixMode(mode) ;
    }

  public :
    /**
     * ムーブコンストラクタ。
     */
    Matrix(Matrix &&status) : m_moved(status.m_moved)
    {
      status.m_moved = true ;
    }

    /**
     * デストラクタ。
     */
    ~Matrix()
    {
      // 移動済みでない場合は初期値に戻す
      if (!m_moved) {
        // matrix modeを変更
        GLint mode ;
        glGetIntegerv(GL_MATRIX_MODE, &mode) ;
        if (mode != SetterConst) glMatrixMode(SetterConst) ;

        // 退避した値に戻す
        glPopMatrix() ;

        // matrix modeを元に戻す
        if (mode != SetterConst) glMatrixMode(mode) ;
      }
    }

    /**
     * 行列値を設定する。
     *
     * @param  enabled 新しい行列値。
     * @return         オブジェクト自身。
     */
    template <class T>
    Matrix &operator = (const ::Matrix<4, 4, T> &mat)
    {
      // matrix modeを変更
      GLint mode ;
      glGetIntegerv(GL_MATRIX_MODE, &mode) ;
      if (mode != SetterConst) glMatrixMode(SetterConst) ;

      // 行列値の設定
      MatrixFunctions<T>::set(mat) ;

      // matrix modeを元に戻す
      if (mode != SetterConst) glMatrixMode(mode) ;

      return *this ;
    }

    /**
     * 行列値を取得。
     *
     * @return 現在の行列値。
     */
    template <class T>
    operator ::Matrix<4, 4, T>() const
    {
      ::Matrix<4, 4, T> mat ;
      MatrixFunctions<T>::get(GetterConst, mat) ;

      return mat ;
    }

    /**
     * 行列を乗算する。
     */
    template <class T>
    Matrix &operator *= (const ::Matrix<4, 4, T> &mat)
    {
      // matrix modeを変更
      GLint mode ;
      glGetIntegerv(GL_MATRIX_MODE, &mode) ;
      if (mode != SetterConst) glMatrixMode(SetterConst) ;

      // 乗算
      MatrixFunctions<T>::mult(mat) ;

      // matrix modeを元に戻す
      if (mode != SetterConst) glMatrixMode(mode) ;

      return *this ;
    }
    
  private :
    bool m_moved ;
  } ;
}

#endif
