#ifndef gl__Texture_h
#define gl__Texture_h

#include "../glall.h"
#include "../flag_util.h"
#include <stddef.h>
#include <cstring>

namespace gl
{
  class TextureObject ;

  /**
   * テクスチャの種類。
   */
  enum class TextureType : GLenum
  {
    /**
     * 縦横のサイズが2の冪乗で、パラメータ範囲が縦横(0.0〜1.0)になるようなテクスチャ。
     */
    Normal = GL_TEXTURE_2D ,
    /**
     * 縦横のサイズが任意で、パラメータ範囲がピクセル単位になるようなテクスチャ。
     */
    Rect   = GL_TEXTURE_RECTANGLE ,
  } ;  

  /**
   * ピクセルデータのフォーマット種類。
   */
  enum class PixelFormat : GLenum
  {
    Red  = GL_RED ,
    RG   = GL_RG ,
    RGB  = GL_RGB ,
    BGR  = GL_BGR ,
    RGBA = GL_RGBA ,
    BGRA = GL_BGRA ,
    Depth           = GL_DEPTH_COMPONENT ,
    DepthAndStencil = GL_DEPTH_STENCIL ,
  } ;

  /**
   * ピクセルデータの種類。
   */
  enum class PixelDataType : GLenum
  {
    UByte = GL_UNSIGNED_BYTE ,
    UShort = GL_UNSIGNED_SHORT ,
    Short = GL_SHORT ,
    UInt = GL_UNSIGNED_INT ,
    Float = GL_FLOAT ,
    UByte_3_3_2 = GL_UNSIGNED_BYTE_3_3_2 ,
    UByte_2_3_3_Rev = GL_UNSIGNED_BYTE_2_3_3_REV ,
    UShort_5_6_5 = GL_UNSIGNED_SHORT_5_6_5 ,
    UShort_5_6_5_Rev = GL_UNSIGNED_SHORT_5_6_5_REV ,
    UShort_4_4_4_4 = GL_UNSIGNED_SHORT_4_4_4_4 ,
    UShort_4_4_4_4_Rev = GL_UNSIGNED_SHORT_4_4_4_4_REV ,
    UShort_5_5_5_1 = GL_UNSIGNED_SHORT_5_5_5_1 ,
    UShort_1_5_5_5_Rev = GL_UNSIGNED_SHORT_1_5_5_5_REV ,
    UInt_8_8_8_8 = GL_UNSIGNED_INT_8_8_8_8,
    UInt_8_8_8_8_Rev = GL_UNSIGNED_INT_8_8_8_8_REV,
    UInt_10_10_10_2 = GL_UNSIGNED_INT_10_10_10_2,
    UInt_2_10_10_10_Rev = GL_UNSIGNED_INT_2_10_10_10_REV,
  } ;

  /**
   * 内部フォーマットの種類。一部省略。
   */
  enum class InternalFormat : GLenum
  {
    Red  = GL_RED ,
    RG   = GL_RG ,
    RGB  = GL_RGB ,
    RGBA = GL_RGBA ,
    Depth           = GL_DEPTH_COMPONENT ,
    DepthAndStencil = GL_DEPTH_STENCIL ,

    Red_8 = GL_R8 ,
    Red_16 = GL_R16 ,
    RG_8_8 = GL_RG8 ,
    RG_16_16 = GL_RG16 ,
    RGB_3_3_2 = GL_R3_G3_B2 ,
    RGB_4_4_4 = GL_RGB4 ,
    RGB_5_5_5 = GL_RGB5 ,
    RGB_8_8_8 = GL_RGB8 ,
    RGB_10_10_10 = GL_RGB10 ,
    RGB_12_12_12 = GL_RGB12 ,
    RGBA_2_2_2_2 = GL_RGBA2 ,
    RGBA_4_4_4_4 = GL_RGBA4 ,
    RGBA_5_5_5_1 = GL_RGB5_A1 ,
    RGBA_8_8_8_8 = GL_RGBA8 ,
    RGBA_10_10_10_2 = GL_RGB10_A2 ,
    RGBA_12_12_12_12 = GL_RGBA12 ,
    RGBA_16_16_16_16 = GL_RGBA16 ,
  } ;

  /**
   * PixelFormat→既定のInternalFormatを求める構造体。
   */
  template <PixelFormat format>
  struct DefaultInternalFormat { static const InternalFormat value = static_cast<InternalFormat>(format) ; } ;
  template <> struct DefaultInternalFormat<PixelFormat::BGR>  { static const InternalFormat value = InternalFormat::RGB ; } ;
  template <> struct DefaultInternalFormat<PixelFormat::BGRA> { static const InternalFormat value = InternalFormat::RGBA ; } ;

  /**
   * PixelDataType→型を求める構造体。
   */
  template <PixelDataType T> struct PixelValue {} ;
  template <> struct PixelValue<PixelDataType::UByte>               { typedef GLubyte  type ; static const bool packed = false ; } ;
  template <> struct PixelValue<PixelDataType::UShort>              { typedef GLushort type ; static const bool packed = false ; } ;
  template <> struct PixelValue<PixelDataType::Short>               { typedef GLshort  type ; static const bool packed = false ; } ;
  template <> struct PixelValue<PixelDataType::UInt>                { typedef GLuint   type ; static const bool packed = false ; } ;
  template <> struct PixelValue<PixelDataType::Float>               { typedef GLfloat  type ; static const bool packed = false ; } ;
  template <> struct PixelValue<PixelDataType::UByte_3_3_2>         { typedef GLubyte  type ; static const bool packed = true  ; } ;
  template <> struct PixelValue<PixelDataType::UByte_2_3_3_Rev>     { typedef GLubyte  type ; static const bool packed = true  ; } ;
  template <> struct PixelValue<PixelDataType::UShort_5_6_5>        { typedef GLushort type ; static const bool packed = true  ; } ;
  template <> struct PixelValue<PixelDataType::UShort_5_6_5_Rev>    { typedef GLushort type ; static const bool packed = true  ; } ;
  template <> struct PixelValue<PixelDataType::UShort_4_4_4_4>      { typedef GLushort type ; static const bool packed = true  ; } ;
  template <> struct PixelValue<PixelDataType::UShort_4_4_4_4_Rev>  { typedef GLushort type ; static const bool packed = true  ; } ;
  template <> struct PixelValue<PixelDataType::UShort_5_5_5_1>      { typedef GLushort type ; static const bool packed = true  ; } ;
  template <> struct PixelValue<PixelDataType::UShort_1_5_5_5_Rev>  { typedef GLushort type ; static const bool packed = true  ; } ;
  template <> struct PixelValue<PixelDataType::UInt_8_8_8_8>        { typedef GLuint   type ; static const bool packed = true  ; } ;
  template <> struct PixelValue<PixelDataType::UInt_8_8_8_8_Rev>    { typedef GLuint   type ; static const bool packed = true  ; } ;
  template <> struct PixelValue<PixelDataType::UInt_10_10_10_2>     { typedef GLuint   type ; static const bool packed = true  ; } ;
  template <> struct PixelValue<PixelDataType::UInt_2_10_10_10_Rev> { typedef GLuint   type ; static const bool packed = true  ; } ;

  /**
   * ピクセルあたりのデータ数。
   */
  template <PixelFormat format> struct PixelComponents {} ;
  template <> struct PixelComponents<PixelFormat::Red>  { static const size_t count = 1 ; } ;
  template <> struct PixelComponents<PixelFormat::RG>   { static const size_t count = 2 ; } ;
  template <> struct PixelComponents<PixelFormat::RGB>  { static const size_t count = 3 ; } ;
  template <> struct PixelComponents<PixelFormat::BGR>  { static const size_t count = 3 ; } ;
  template <> struct PixelComponents<PixelFormat::RGBA> { static const size_t count = 4 ; } ;
  template <> struct PixelComponents<PixelFormat::BGRA> { static const size_t count = 4 ; } ;
  template <> struct PixelComponents<PixelFormat::Depth>           { static const size_t count = 1 ; } ;
  template <> struct PixelComponents<PixelFormat::DepthAndStencil> { static const size_t count = 2 ; } ;

  /**
   * ラップ方法。
   */
  enum class TextureWrap : GLint
  {
    Clamp  = GL_CLAMP ,
    ClampToEdge = GL_CLAMP_TO_EDGE ,
    ClampToBorder = GL_CLAMP_TO_BORDER ,
    Repeat = GL_REPEAT ,
    MirroredRepeat = GL_MIRRORED_REPEAT ,
  } ;

  /**
   * 縮小関数。
   */
  enum class TextureMinFilter : GLint
  {
    Nearest = GL_NEAREST ,
    Linear  = GL_LINEAR ,
  } ;

  /**
   * 拡大関数。
   */
  enum class TextureMagFilter : GLint
  {
    Nearest = GL_NEAREST ,
    Linear  = GL_LINEAR ,
  } ;

  /**
   * コンテキスト非依存のテクスチャデータ。
   */
  class Texture
  {
    friend class TextureObject ;
    friend class TextureRenderer ;

  public :
    /**
     * Textureクラスで変更された情報の列挙。
     */
    enum class ChangedTextureData : unsigned int
    {
      None        = 0x0000 ,
      Image       = 0x0001 ,
      BorderColor = 0x0002 ,
      MagFilter   = 0x0004 ,
      MinFilter   = 0x0008 ,
      WrapS       = 0x0010 ,
      WrapT       = 0x0020 ,

      All         = static_cast<unsigned int>(-1),
    } ;

  public :
    /**
     * コンストラクタ。
     *
     * @param target         テクスチャの種類。
     */
    Texture(TextureType target) : m_target(target) {}

    /**
     * コピー・コンストラクタ。
     */
    Texture(const Texture &another) ;

    /**
     * デストラクタ。
     */
    ~Texture()
    {
      notifyDelete() ;
      if (m_buffer) delete[] m_buffer ;
    }

    /**
     * 代入。
     */
    Texture &operator =(const Texture &another) ;

    /**
     * テクスチャの種類を取得。
     */
    TextureType getTarget() const { return m_target; }

    /**
     * 拡大関数を取得。
     */
    TextureMagFilter getMagFilter() const { return m_mag_filter ; }

    /**
     * 縮小関数を取得。
     */
    TextureMinFilter getMinFilter() const { return m_min_filter ; }

    /**
     * S方向のラップ方法を取得。
     */
    TextureWrap getWrapS() const { return m_wrap_s ; }

    /**
     * T方向のラップ方法を取得。
     */
    TextureWrap getWrapT() const { return m_wrap_t ; }

    /**
     * 枠線色を取得。
     */
    const GLfloat (&getBorderColor() const)[4] { return m_border_color ; }

    /**
     * 画像の幅を取得。
     */
    GLsizei getWidth() const { return m_width ; }

    /**
     * 画像の高さを取得。
     */
    GLsizei getHeight() const { return m_height ; }

    /**
     * 拡大関数を設定。
     */
    void setMagFilter(TextureMagFilter filter)
    {
      if (m_mag_filter != filter) {
        notifyChange(ChangedTextureData::MagFilter) ;
        m_mag_filter = filter ;
      }
    }

    /**
     * 縮小関数を設定。
     */
    void setMinFilter(TextureMinFilter filter)
    {
      if (m_min_filter != filter) {
        notifyChange(ChangedTextureData::MinFilter) ;
        m_min_filter = filter ;
      }
    }

    /**
     * S方向のラップ方法を設定。
     */
    void setWrapS(TextureWrap wrap)
    {
      if (m_wrap_s != wrap) {
        notifyChange(ChangedTextureData::WrapS) ;
        m_wrap_s = wrap ;
      }
    }

    /**
     * T方向のラップ方法を設定。
     */
    void setWrapT(TextureWrap wrap)
    {
      if (m_wrap_t != wrap) {
        notifyChange(ChangedTextureData::WrapT) ;
        m_wrap_t = wrap ;
      }
    }

    /**
     * 枠線色を設定。
     */
    void setBorderColor(const GLfloat *rgba)
    {
      for (GLfloat *p = m_border_color, *pEnd = p + 4 ; p < pEnd ; ++p, ++rgba) {
        if (*p != *rgba) {
          notifyChange(ChangedTextureData::BorderColor) ;
          for ( ; p < pEnd ; ++p, ++rgba) *p = *rgba ;
          return ;
        }
      }
    }

    /**
     * 画像を設定。
     *
     * @param format         ピクセルデータのフォーマット。
     * @param type           ピクセルデータの種類。
     * @param internalFormat 内部フォーマット。
     */
    template <PixelFormat format = PixelFormat::RGBA, PixelDataType type = PixelDataType::UByte,
              InternalFormat internalFormat = DefaultInternalFormat<format>::value>
    void setBuffer(GLsizei width, GLsizei height, const typename PixelValue<type>::type *buffer, GLint border = 0)
    {
      // ピクセルあたりのデータ数。
      const size_t value_size = PixelValue<type>::packed ? 1 : PixelComponents<format>::count ;

      notifyChange(ChangedTextureData::Image) ;

      m_internalFormat = internalFormat ;
      m_format = format ;
      m_type = type ;
      m_border = border ;

      m_width = width ;
      m_height = height ;

      size_t new_size = width * height * value_size * sizeof(value_size) ;
      if (!m_buffer) {
        // メモリを未確保なので確保
        m_buffer_size = new_size ;
        m_buffer = new GLubyte[m_buffer_size] ;
      }
      else {
        if (m_buffer_size < new_size) {
          // 新サイズの方が大きいのでメモリを再確保
          m_buffer_size = new_size ;
          delete[] m_buffer ;
          m_buffer = new GLubyte[m_buffer_size] ;
        }
      }
      std::memcpy(m_buffer, buffer, new_size) ;
    }

  private :
    /**
     * 各コンテキストに変更を通知。
     */
    void notifyChange(ChangedTextureData data) ;

    /**
     * 各コンテキストに削除を通知。
     */
    void notifyDelete() ;

  private :
    TextureType m_target ;

    GLsizei m_width = 0 ;
    GLsizei m_height = 0 ;
    InternalFormat m_internalFormat = InternalFormat::RGBA ;
    PixelFormat m_format = PixelFormat::RGBA ;
    PixelDataType m_type = PixelDataType::UByte ;
    GLint m_border = 0 ;
    GLubyte *m_buffer = nullptr ;
    size_t m_buffer_size = 0 ;

    TextureMagFilter m_mag_filter = TextureMagFilter::Nearest ;
    TextureMinFilter m_min_filter = TextureMinFilter::Nearest ;
    TextureWrap m_wrap_s = TextureWrap::Repeat ;
    TextureWrap m_wrap_t = TextureWrap::Repeat ;
    GLfloat m_border_color[4] = {0} ;

  private :
    /**
     * テクスチャプロキシ用の定数を求める。
     */
    GLenum getTextureProxy() const
    {
      switch (m_target) {
        case TextureType::Normal : return GL_PROXY_TEXTURE_2D ;
        case TextureType::Rect   : return GL_PROXY_TEXTURE_RECTANGLE ;
        default : return 0 ; // ERROR
      }
    }

    /**
     * バインドされているテクスチャを取得するための定数を求める。
     */
    GLenum getBindingParam() const
    {
      switch (m_target) {
        case TextureType::Normal : return GL_TEXTURE_BINDING_2D ;
        case TextureType::Rect   : return GL_TEXTURE_BINDING_RECTANGLE ;
        default : return 0 ; // ERROR
      }
    }
  } ;
}

define_flag_operators(gl::Texture::ChangedTextureData) ;

#endif
