#include "Window.h"
#include "../glall.h"

// static変数定義
std::map<int, glut::Window*> glut::Window::m_id_map ;

glut::Window::Window(const char *name)
  : m_window_id(0), m_context(this)
{
  // GLUTを用いてウィンドウを作成
  m_window_id = glutCreateWindow(name) ;
  if (!m_window_id) throw std::exception() ; // 作成に失敗

  // ウィンドウID→ウィンドウの辞書の登録
  m_id_map.insert({m_window_id, this}) ;

  // 各種関数を登録
  initialize() ;
}

glut::Window::~Window()
{
  if (!m_window_id) return ;

  // GLUTを用いてウィンドウを破棄
  glutDestroyWindow(m_window_id) ;

  // ウィンドウID→ウィンドウの辞書の登録解除
  m_id_map.erase(m_window_id) ;
}

void glut::Window::refresh()
{
  glutSetWindow(m_window_id) ;
  glutPostRedisplay() ;
}

void glut::Window::moveTo(int x, int y)
{
  glutSetWindow(m_window_id) ;
  glutPositionWindow(x, y) ;
}

void glut::Window::resizeTo(int width, int height)
{
  glutSetWindow(m_window_id) ;
  glutReshapeWindow(width, height) ;
}

void glut::Window::setFullScreen()
{
  glutSetWindow(m_window_id) ;
  glutFullScreen() ;
}

void glut::Window::initialize()
{
  // 各種処理を登録
  glutDisplayFunc(displayFunc) ;
  glutReshapeFunc(reshapeFunc) ;
  glutKeyboardFunc(keyboardFunc) ;
  glutSpecialFunc(specialFunc) ;
  glutMouseFunc(mouseFunc) ;
  glutMotionFunc(motionFunc) ;
  glutMotionFunc(passiveMotionFunc) ;
  glutEntryFunc(entryFunc) ;
}

template <class ... TParams>
void glut::Window::execute(void (Window::*method)(gl::Context &, TParams ...), TParams ... parameters)
{
  // 現在のウィンドウIDからウィンドウを取得
  auto it = m_id_map.find(glutGetWindow()) ;
  if (it == m_id_map.end()) return ;

  // コンテキストを有効化
  it->second->m_context.awake() ;

  // ウィンドウに対して、指定したメソッドを呼び出す
  (it->second->*method)(it->second->m_context, parameters ...) ;
}

void glut::Window::displayFunc()
{
  execute(&Window::display) ;
}

void glut::Window::reshapeFunc(int width, int height)
{
  execute(&Window::reshape, width, height) ;
}

void glut::Window::keyboardFunc(unsigned char key, int x, int y)
{
  execute(&Window::keyboard, key, x, y) ;
}

void glut::Window::specialFunc(int key, int x, int y)
{
  execute(&Window::special, static_cast<SpecialKey>(key), x, y) ;
}

void glut::Window::mouseFunc(int button, int state, int x, int y)
{
  execute(&Window::mouse, static_cast<MouseButton>(button), static_cast<ButtonState>(state), x, y) ;
}

void glut::Window::motionFunc(int x, int y)
{
  execute(&Window::motion, ButtonState::Down, x, y) ;
}

void glut::Window::passiveMotionFunc(int x, int y)
{
  execute(&Window::motion, ButtonState::Up, x, y) ;
}

void glut::Window::entryFunc(int state)
{
  execute(&Window::entry, static_cast<MouseEntry>(state)) ;
}
