#ifndef glut__Window_h
#define glut__Window_h

#include <map>
#include "../noncopyable.h"
#include "Context.h"
#include "InputConstants.h"

namespace glut
{
  /**
   * GLUTウィンドウのラッパークラス。
   *
   * ウィンドウはコピー不可（参照のみ可能とする）。
   */
  class Window : noncopyable<Window>
  {
  public :
    /**
     * コンストラクタ。
     *
     * @param name ウィンドウの名前。
     */
    Window(const char *name) ;
    virtual ~Window() ;

    /**
     * ウィンドウのIDの取得。
     *
     * @return ウィンドウのID。
     */
    int getID() const { return m_window_id ; }

    /**
     * ウィンドウの再描画を要求。
     */
    void refresh() ;

    /**
     * ウィンドウ位置の変更。
     *
     * @param x 変更後のウィンドウのX位置。
     * @param y 変更後のウィンドウのY位置。
     */
    void moveTo(int x, int y) ;

    /**
     * ウィンドウサイズの変更。
     *
     * @param width  変更後のウィンドウの幅。
     * @param height 変更後のウィンドウの高さ。
     */
    void resizeTo(int width, int height) ;

    /**
     * ウィンドウのフルスクリーン化。
     */
    void setFullScreen() ;

  protected :
    /**
     * 各種の関数を登録。
     */
    void initialize() ;

    /**
     * ウィンドウに対する描画処理。
     *
     * @param context OpenGLコンテキスト
     */
    virtual void display(gl::Context &context) {}

    /**
     * ウィンドウリサイズ時の処理。
     *
     * @param context OpenGLコンテキスト。
     * @param width   変更後のウィンドウ幅。
     * @param height  変更後のウィンドウ高さ。
     */
    virtual void reshape(gl::Context &context, int width, int height) {}

    /**
     * キー押下時の処理。
     *
     * @param context OpenGLコンテキスト。
     * @param key     押下されたキー。
     * @param x       キーが押下された時のマウスのX座標。
     * @param y       キーが押下された時のマウスのY座標。
     */
    virtual void keyboard(gl::Context &context, unsigned char key, int x, int y) {}

    /**
     * 特殊キー押下時の処理。
     *
     * @param context OpenGLコンテキスト。
     * @param key     押下された特殊キー。
     * @param x       特殊キーが押下された時のマウスのX座標。
     * @param y       特殊キーが押下された時のマウスのY座標。
     */
    virtual void special(gl::Context &context, SpecialKey key, int x, int y) {}

    /**
     * マウスボタンの処理。
     *
     * @param context OpenGLコンテキスト。
     * @param button  マウスボタン。
     * @param state   ボタンの状態。
     * @param x       マウスのX座標。
     * @param y       マウスのY座標。
     */
    virtual void mouse(gl::Context &context, MouseButton key, ButtonState state, int x, int y) {}

    /**
     * マウス移動の処理。
     *
     * @param context OpenGLコンテキスト。
     * @param state   ボタンの状態。
     * @param x       マウスのX座標。
     * @param y       マウスのY座標。
     */
    virtual void motion(gl::Context &context, ButtonState state, int x, int y) {}

    /**
     * マウス出入りに対する処理。
     *
     * @param context OpenGLコンテキスト。
     * @param state   マウスの出入り。
     */
    virtual void entry(gl::Context &context, MouseEntry state) {}
    

  private :
    /**
     * ウィンドウのID。
     */
    int m_window_id ;
    /**
     * ウィンドウの持つOpenGLコンテキスト。
     */
    Context m_context ;

  private :
    /**
     * コールバックを各々のウィンドウの指定メソッドに割り当てる処理。
     */
    template <class ... TParams>
    static void execute(void (Window::*method)(gl::Context &, TParams ...), TParams ... parameters) ;

    /**
     * 描画処理を各々のウィンドウのdisplay()に割り当てる処理。
     */
    static void displayFunc() ;

    /**
     * ウィンドウリサイズ時の処理を各々のウィンドウのreshape()に割り当てる処理。
     *
     * @param width  変更後のウィンドウ幅。
     * @param height 変更後のウィンドウ高さ。
     */
    static void reshapeFunc(int width, int height) ;

    /**
     * キー押下時の処理を各々のウィンドウのkeyboard()に割り当てる処理。
     *
     * @param key     押下されたキー。
     * @param x       キーが押下された時のマウスのX座標。
     * @param y       キーが押下された時のマウスのY座標。
     */
    static void keyboardFunc(unsigned char key, int x, int y) ;

    /**
     * 特殊キー押下時の処理を各々のウィンドウのspecial()に割り当てる処理。
     *
     * @param key     押下された特殊キー。
     * @param x       特殊キーが押下された時のマウスのX座標。
     * @param y       特殊キーが押下された時のマウスのY座標。
     */
    static void specialFunc(int key, int x, int y) ;

    /**
     * マウスボタンの処理を各々のウィンドウのmouse()に割り当てる処理。
     *
     * @param button  マウスボタン。
     * @param state   ボタンの状態。
     * @param x       マウスのX座標。
     * @param y       マウスのY座標。
     */
    static void mouseFunc(int button, int state, int x, int y) ;

    /**
     * ボタン押下時のマウス移動の処理を各々のウィンドウのmotion()に割り当てる処理。
     *
     * @param x       マウスのX座標。
     * @param y       マウスのY座標。
     */
    static void motionFunc(int x, int y) ;
    
    /**
     * ボタン非押下時のマウス移動の処理を各々のウィンドウのmotion()に割り当てる処理。
     *
     * @param x       マウスのX座標。
     * @param y       マウスのY座標。
     */
    static void passiveMotionFunc(int x, int y) ;

    /**
     * マウス出入りに対する処理を各々のウィンドウのentry()に割り当てる処理。
     *
     * @param state   マウスの出入り。
     */
    static void entryFunc(int state) ;

    /**
     * ウィンドウのIDからWindowオブジェクトを求めるための辞書。
     */
    static std::map<int, Window*> m_id_map ;
  } ;
}

#endif
